(**************************************************************)
(*   The Strong Completeness of Partial Monoidal Boolean BI   *)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import List.

Require Import predicate.
Require Import enum.
Require Import enumerations.
Require Import carrier.
Require Import finite.

Require Import words.

Notation "x ~~ y" := ((x,y) : _) (at level 40, no associativity).

Section constraints.

  Set Implicit Arguments.

  Variable letter : Type.

  Definition constraint := (word letter * word letter)%type.
  
  Section enum_constraint.

    Hypothesis enum_letter : enumeration letter.
    
    Definition enum_constraint : enumeration constraint.
    Proof.
      generalize (enum_word enum_letter).
      intros; apply enum_prod; auto.
    Qed.
    
    Proposition constraint_eq_dec (a b : constraint) : { a = b } + { a <> b }.
    Proof. apply enumeration_dec, enum_constraint. Qed.
    
  End enum_constraint.

  (* List of labels is a (non-commutative) monoid *)

  Definition language (R : _ -> Prop) (x : word letter) := exists m, exists n, R (m~~n) /\ (subword x m \/ subword x n).
  Definition alphabet (R : _ -> Prop) (a : letter) := language R (l a).

  Section LanguageAlphabet.

    Proposition language_inc R1 R2 : R1 inc R2 -> language R1 inc language R2.
    Proof. intros ? ? ( m & n & ? & ? ); exists m; exists n; split; auto. Qed.
    
    Proposition alphabet_inc R1 R2 : R1 inc R2 -> alphabet R1 inc alphabet R2.
    Proof. intros ? ?. unfold alphabet; apply language_inc; auto. Qed.

    Hint Resolve language_inc alphabet_inc.
    
    Proposition language_id R1 R2 : R1 ~ R2 -> language R1 ~ language R2.
    Proof. intros []; auto. Qed.
    
    Proposition alphabet_id R1 R2 : R1 ~ R2 -> alphabet R1 ~ alphabet R2.
    Proof. intros []; auto. Qed.
    
    Hint Resolve language_id alphabet_id.
    
    Proposition alphabet_empty : alphabet empty ~ empty.
    Proof. split; auto; intros ? ( ? & ? & Hx & _  ); contradiction Hx. Qed.
    
    Proposition alphabet_sg m n : alphabet (sg (m~~n)) ~ carrier m cup carrier n.
    Proof.
      split.
      
      intros a [ m' [ n' [ H0 H1 ] ] ].
      apply in_sg_eq in H0. injection H0. clear H0. intros H H'.
      rewrite H, H'; clear m n H H'.
      destruct H1 as [ H1 | H1 ]; apply subletter_inc_carrier in H1; 
        [ apply bin_union_inc_r1 | apply bin_union_inc_r2]; auto.
      
      apply bin_union_inc_l;
        intros a Ha; apply carrier_inc_subletter in Ha; 
        exists m; exists n; split; ( apply in_sg_reflexive || trivial );
        [ left | right ]; trivial.
    Qed.
    
    Hint Immediate alphabet_sg.

    Implicit Type R : constraint -> Prop.
    
    Proposition alphabet_R_inc_l R m n : R (m ~~ n) -> carrier m inc alphabet R.
    Proof.
      intros.
      apply inc_transitive with (B := alphabet (sg (m ~~ n))); auto.
      apply inc_transitive with (2 := proj2 (alphabet_sg _ _)); auto.
    Qed.
    
    Proposition alphabet_R_inc_r R m n : R (m ~~ n) -> carrier n inc alphabet R.
    Proof.
      intros.
      apply inc_transitive with (B := alphabet (sg (m ~~ n))); auto.
      apply inc_transitive with (2 := proj2 (alphabet_sg _ _)); auto.
    Qed.

    Proposition alphabet_sg_eps : alphabet (sg (eps~~eps)) inc empty.
    Proof.
      intros a ( m & n & H0 & H1 ).
      apply in_sg_eq in H0; injection H0; clear H0; intros H H'.
      rewrite <- H, <- H' in H1; clear H H'.
      destruct H1 as [ H1 | H1 ]; destruct H1 as [x H]; apply weq_length in H; simpl in H; discriminate H.
    Qed.
    
    Proposition alphabet_sg_lab a : alphabet (sg (l a ~~ l a)) ~ sg a.
    Proof.
      apply id_transitive with (1 := alphabet_sg _ _).
      apply id_transitive with (1 := bin_union_idem _ _).
      split; auto.
    Qed.
    
    Proposition alphabet_cup R1 R2 : alphabet (R1 cup R2) ~ alphabet R1 cup alphabet R2.
    Proof.
      split; auto.
      intros a ( m & n & Hmn & Hmn' ).
      apply bin_union_sg in Hmn. 
      destruct Hmn as [ Hmn | Hmn ]; [ apply bin_union_inc_r1 | apply bin_union_inc_r2 ]; exists m; exists n; tauto.
    Qed.
    
    Let alphabet_nil : alphabet (carrier nil) ~ empty.
    Proof.
      split; auto.
      intros x [ a [ b [ Hx _ ] ] ].
      simpl in Hx. contradiction Hx.
    Qed.
    
    Let alphabet_cons x y C : alphabet (carrier ((x~~y)::C)) ~ carrier x cup carrier y cup alphabet (carrier C).
    Proof.
      apply id_transitive with (1 := alphabet_id (carrier_cons _ _ _)).
      apply id_transitive with (1 := alphabet_cup _ _).
      auto.
    Qed.

    Proposition alphabet_finite R : finite R -> finite (alphabet R).
    Proof.
      intros [ C HC ]. 
      apply (finite_id _ _ _ (alphabet_id HC)).
      clear R HC.
      induction C as [ | (x,y) C ].
      apply (finite_id _ empty); auto. 
      apply (finite_id _ _ _ (id_symetric _ _ _ (alphabet_cons _ _ _))); auto.
    Qed.

  End LanguageAlphabet.

End constraints.

Hint Resolve alphabet_empty alphabet_sg alphabet_sg_lab alphabet_sg_eps alphabet_cup.
Hint Resolve alphabet_R_inc_l alphabet_R_inc_r.
Hint Resolve language_inc alphabet_inc language_id alphabet_id alphabet_finite.

