(**************************************************************)
(*   The Strong Completeness of Partial Monoidal Boolean BI   *)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import Arith.
Require Import MinMax.

Require Import predicate.
Require Import enum.
Require Import enumerations.

Parameter Var : Type.
Parameter enum_var : enumeration Var.

Inductive BI_Cst : Type := BI_Unit.
Inductive BI_UnOp : Type := BI_Not.
Inductive BI_BinOp : Type := BI_Star | BI_Wand | BI_And.

Inductive formula : Type :=
  | Form_var (_ : Var)
  | Form_cst (_ : BI_Cst)
  | Form_unop (_ : BI_UnOp) (_ : formula)
  | Form_binop (_ : BI_BinOp) (_ _ : formula).

Section FormEnum.

  Let enum_cst : enumeration BI_Cst.
  Proof.
    exists (fun n => match n with _ => BI_Unit end)
           (fun x => match x with _ => 0 end);
    induction x; auto.
  Qed.

  Definition enum_unop : enumeration BI_UnOp.
  Proof.
    exists (fun _ => BI_Not) 
           (fun _ => 0);
    induction x; auto.
  Qed.

  Definition enum_binop : enumeration BI_BinOp.
  Proof.
    exists (fun n => match n with | 0 => BI_Star | 1 => BI_Wand | _ => BI_And end)
           (fun x => match x with | BI_Star => 0 | BI_Wand => 1 | BI_And => 2 end);
    induction x; auto.
  Qed.

  Let max_prop a b c : max a b <= c -> a<=c /\ b<= c.
  Proof. intros H; split; apply le_trans with (2 := H); auto with arith. Qed.

  Let form_height : formula -> nat := 
    let fix loop f { struct f } := match f with
      | Form_var _ => 0
      | Form_cst _ => 0
      | Form_unop _ f => S (loop f)
      | Form_binop _ f g => S (max (loop f) (loop g))
    end in loop. 

  Let form_bounded n t := form_height t <= n.

  Let form_enum_rec n : partial_enumeration (form_bounded n).
  Proof.
    induction n.

    set (eO := enum_sum _ _ enum_var enum_cst).
    destruct eO as [ dO rO sO ].
    set (dX := fun n => match dO n with inl v => Form_var v | inr c => Form_cst c end).
    set (rX := fun f => match f with Form_var v => rO (inl _ v) | Form_cst c => rO (inr _ c) | _ => O end).
    exists dX rX. induction x. 
    intros H; unfold dX; simpl; rewrite -> sO; auto.
    intros H; unfold dX; simpl; rewrite -> sO; auto.
    intros H; contradiction (le_Sn_O _ H).
    intros H; contradiction (le_Sn_O _ H).

    set (eO := enum_sum _ _ enum_var 
                    (enum_sum _ _ enum_cst 
                    (enum_sum _ _ (enum_prod _ _ enum_unop nat_enum) 
                                             (enum_prod _ _ enum_binop
                                             (enum_prod _ _ nat_enum nat_enum))))).
    destruct IHn as [ dHn rHn sHn ].
    destruct eO as [ dO rO sO ].

    set (dX := fun n => match dO n with 
                   | inl v => Form_var v 
                   | inr (inl c) => Form_cst c 
                   | inr (inr (inl (u,n))) => Form_unop u (dHn n)
                   | inr (inr (inr (b,(n1,n2)))) => Form_binop b (dHn n1) (dHn n2) 
              end).
    set (rX := fun f => match f with
                   | Form_var v => rO (inl _ v)
                   | Form_cst c => rO (inr _ (inl _ c))
                   | Form_unop u f => rO (inr _ (inr _ (inl _ (u, rHn f)))) 
                   | Form_binop b f1 f2 => rO (inr _ (inr _ (inr _ (b, (rHn f1, rHn f2)))))
              end).
    exists dX rX. induction x as [ v | [] | ? A | ? A IHA B IHB ].
    intros; unfold dX, rX; rewrite -> sO; auto.
    intros; unfold dX, rX; rewrite -> sO; auto.
    intros H; unfold dX, rX; rewrite -> sO, -> sHn; auto; apply le_S_n, H.
    intros H; red in H; simpl in H; apply le_S_n in H. 
    unfold dX, rX; rewrite -> sO, -> sHn, -> sHn; auto; apply le_trans with (2 := H); auto with arith.
  Qed.

  Definition enum_formula : enumeration formula.
  Proof.
    apply (enum_layers _ form_bounded form_height); auto.
    intros; red; auto.
  Qed.

End FormEnum.

(* equality is decided using the enumeration *)

Definition form_eq_dec : forall x y : formula, { x = y } + { x <> y }.
Proof.
  apply enumeration_dec.
  apply enum_formula.
Qed.

Delimit Scope Formula_scope with formula.
  
Notation "x * y" := (Form_binop BI_Star x y) (at level 40, left associativity) : Formula_scope.
Notation "x '-*' y" := (Form_binop BI_Wand x y) (at level 41, left associativity) : Formula_scope.
Notation "x '&&' y" := (Form_binop BI_And x y) (at level 40, left associativity) : Formula_scope.
Notation " ! x " := (Form_unop BI_Not x) (at level 39, right associativity) : Formula_scope.
Notation "% v" := (Form_var v) (at level 30) : Formula_scope.
Notation "'I'" := (Form_cst BI_Unit) : Formula_scope.

Open Scope Formula_scope.
