(**************************************************************)
(*   The Strong Completeness of Partial Monoidal Boolean BI   *)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import Arith.
Require Import Omega.

Section Defs.

  Variable P : nat -> Prop.

  Definition pnat_dec := forall x, { P x } + { ~ P x }.
  Definition pnat_inf := forall x, { y | x <= y /\ P y }.

End Defs.

Section NextP.

  Variable P : nat -> Prop.
  Variable Pdec : pnat_dec P.
  
  Let nextP_rec n : forall x, P (x+n) -> { y | x <= y 
                                            /\ y <= x+n 
                                            /\ P y 
                                            /\ forall k, x <= k -> P k -> y <= k 
                                         }.
  Proof.
    induction n; intros x Hx.

    exists x.
    cutrewrite (x+0=x) in Hx;
    repeat (split; auto; try omega); omega.
    
    intros; auto.
    destruct (Pdec x) as [ HPx | HPx ].

    exists x.
    repeat (split; auto; try omega).

    generalize (IHn (S x)); clear IHn; intros IHn.
    cutrewrite(x+S n=S x+n) in Hx; try omega.
    apply IHn in Hx. clear IHn.
    destruct Hx as ( y & H1 & H2 & H3 & H4 ).
    exists y.
    repeat (split; auto; try omega).
    intros k Hk Hk'.
    destruct (le_lt_dec k x) as [ H | H ].
    rewrite (le_antisym _ _ Hk H) in HPx.
    contradiction (HPx Hk').
    apply H4; auto.
  Qed.

  Definition pnat_minimize_ab a b : a <= b -> P b -> { m | a <= m /\ P m /\ forall x, a <= x -> P x -> m <= x }.
  Proof.
    intros Hab Hb.
    destruct (nextP_rec (b-a) a) as [ m [ H1 [ H2 [ H3  H4 ] ] ] ]; auto.
    cutrewrite (a+(b-a) = b); auto; omega.
    exists m; auto.
  Qed.

  Definition pnat_minimize_a a : P a -> { m | P m /\ forall x, P x -> m <= x }.
  Proof.
    intros Ha.
    destruct (pnat_minimize_ab _ _ (le_0_n _) Ha) as [ m [ H1 [ H2 H3 ] ] ].
    exists m; split; auto.
    intros x; apply H3; omega.
  Qed.

  Variable Pinf : pnat_inf P.

  Definition pnat_next x : { y | x<=y /\ P y /\ forall k, x <= k -> P k -> y <= k }.
  Proof.
    destruct (Pinf x) as [ y [ Hy1 Hy2 ] ].
    destruct (pnat_minimize_ab _ _ Hy1 Hy2) as [ z [ H1 [ H2 H3 ] ] ].
    exists z; repeat (split; auto).
  Qed.

End NextP.

Section NatDecPred.

  Variable P : nat -> Prop.
  Hypothesis Pdec : pnat_dec P.
  
  Definition pnat_bounded_fall i := forall j, j < i -> P j.

  Definition pnat_bounded_fall_rec i : { j | j < i /\ ~ P j } + { pnat_bounded_fall i }.
  Proof.
    induction i.
    right; intros x C; contradict C; omega.
    destruct (Pdec i) as [ Hi | Hi ].
    destruct IHi as [ (j & H1 & H2) | Hr ].
    left; exists j; split; auto; omega.
    right; intros k Hk.
    destruct (eq_nat_dec k i) as [ H | ].
    rewrite H; auto.
    apply Hr; omega.
    left; exists i; split; auto; omega.
  Qed.
 
  Definition pnat_bounded_fall_dec i : { pnat_bounded_fall i } + { ~ pnat_bounded_fall i }.
  Proof. destruct (pnat_bounded_fall_rec i) as [ (? & ? & ?)  | ]; auto. Qed.

End NatDecPred.

Definition pnat_bounded_fall_rec_not P i : pnat_dec P -> { j | j < i /\ P j } + { forall j, j < i -> ~ P j } .
Proof.
  intros HP.
  set (Q i := ~ P i).
  assert (pnat_dec Q) as HQ.
  unfold Q, pnat_dec.
  intros x; destruct (HP x); [right | left];auto.
  destruct (pnat_bounded_fall_rec _ HQ i) as [ (j & H1 & H2) | ]; auto.
  left. exists j; split; auto.
  unfold Q in H2. destruct (HP j); tauto.
Qed. 

