(**************************************************************)
(*   The Strong Completeness of Partial Monoidal Boolean BI   *)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Reserved Notation " x 'inc' y " (at level 60, no associativity).
Reserved Notation " x 'pair' y " (at level 49, no associativity).
Reserved Notation " x 'cup' y " (at level 49).
Reserved Notation " x 'cap' y " (at level 49).

Section Predicate.

  Variable M : Type.

  Definition Predicate := M -> Prop.

  Definition Emptyset (_ : M) := False.
  Definition Fullset  (_ : M) := True. 
  
  Proposition In_fullset : forall x, Fullset x.
  Proof. red; auto. Qed.

  Definition Included (A B : Predicate) : Prop := forall x, A x -> B x.
  Definition sg (a x : M) := a = x.
  
  Definition Pair (a b : M) x := x = a \/ x = b.
  
  Proposition In_pair_left a b : Pair a b a.
  Proof. left; auto. Qed.
  
  Proposition In_pair_right a b : Pair a b b.
  Proof. right; auto. Qed.

  Notation empty := Emptyset.
  Notation full := Fullset.
  Infix "inc" := Included (at level 60, no associativity).
  Infix "pair" := Pair (at level 49, no associativity).

  Proposition inc_reflexive A : A inc A.
  Proof. intro; auto. Qed.

  Proposition inc_transitive A B C : A inc B -> B inc C -> A inc C.
  Proof. intros ? ? ? ?; auto. Qed.

  Proposition inc_in_transitive (B : Predicate) a C :  B a -> B inc C -> C a.
  Proof. intros ? ?; auto. Qed.

  Proposition any_inc_full A : A inc full.
  Proof. intros ? ?; apply In_fullset. Qed.

  Proposition empty_inc_any A : empty inc A.
  Proof. intros ?; induction 1; auto. Qed. 

  Proposition in_sg a A : sg a inc A -> A a.
  Proof. intros H; apply H; constructor. Qed.

  Proposition sg_inc a (A : Predicate) : A a -> sg a inc A.
  Proof. intros ? ?; induction 1; auto. Qed.

  Proposition in_sg_reflexive a : (sg a) a.
  Proof. red; auto. Qed.

  Proposition in_sg_reflexive' a b : a = b -> sg a b.
  Proof. unfold sg; auto. Qed.

  Proposition in_sg_eq a b : (sg a) b -> a = b.
  Proof. unfold sg; auto. Qed.

  Hint Immediate inc_reflexive empty_inc_any in_sg_reflexive in_sg_reflexive'.
  Hint Resolve sg_inc in_sg_eq.

  Definition Identical (A B : Predicate) := A inc B /\ B inc A.
  Infix "~" := Identical (at level 60, no associativity).

  Proposition id_reflexive A : A ~ A.
  Proof. split; auto. Qed.

  Proposition id_symetric A B : A ~ B -> B ~ A.
  Proof. intros [ ]; split; auto. Qed.

  Proposition id_transitive A B C : A ~ B -> B ~ C -> A ~ C.
  Proof. intros [ ? ? ] [ ? ? ]; split; apply inc_transitive with (B := B); auto. Qed.

  Proposition inc_antisym A B : A inc B -> B inc A -> A ~ B.
  Proof. split; auto. Qed.

  (* indexed (& infinite) intersection *)

  Definition inter {X} (phi : X -> Predicate) m := forall x, phi x m.

  Proposition In_intersection X (phi : X -> Predicate) m : (forall x, phi x m) -> inter phi m.
  Proof. auto. Qed.

  Definition union {X} (phi : X -> Predicate) m := exists x, phi x m.

  Proposition In_union X (phi : X -> Predicate) m x : phi x m -> union phi m.
  Proof. exists x; auto. Qed.

  Proposition inter_inc_l X (phi : X -> Predicate) x : inter phi inc phi x.
  Proof. intros ? H; apply H. Qed.

  Hint Immediate inter_inc_l.

  Proposition inter_inc_l' X (phi : X -> Predicate) A x : A inc inter phi -> A inc phi x.
  Proof. intros H; apply inc_transitive with (1:=H); auto. Qed.

  Proposition inter_inc_r X (phi : X -> Predicate) A : (forall x, A inc phi x) -> A inc inter phi.
  Proof. intros H ? ? ?; apply H; auto. Qed.

  Hint Resolve inter_inc_l' inter_inc_r.

  Theorem inter_glb X (phi : X -> Predicate) A : A inc inter phi <-> (forall x, A inc phi x).
  Proof. split; intros; eauto. Qed.

  Proposition union_inc_r X (phi : X -> Predicate) x: phi x inc union phi.
  Proof. intros ? ?; apply (In_union _ _ _ x); auto. Qed.

  Hint Immediate union_inc_r.

  Proposition union_inc_r' X (phi : X -> Predicate) A x : union phi inc A -> phi x inc A.
  Proof. intros H; apply inc_transitive with (2 := H); auto. Qed.

  Theorem union_sg X (phi : X -> Predicate) : forall a, union phi a -> exists x, phi x a.
  Proof. intros; auto. Qed.

  Proposition union_inc_l X (phi : X -> Predicate) A : (forall x, phi x inc A) -> union phi inc A.
  Proof. intros H ? Hx; destruct (union_sg _ _ _ Hx) as [ ? Hk]; apply (H _ _ Hk). Qed.

  Hint Resolve union_inc_r' union_inc_l.

  Theorem union_lub X (phi : X -> Predicate) A : union phi inc A <-> (forall x, phi x inc A).
  Proof. split; intros; eauto. Qed.

  Let bool_map A B x : Predicate := match x with true => A | false => B end. 

  Definition bin_inter A B := inter (bool_map A B).
  Definition bin_union A B := union (bool_map A B).

  Infix "cup" := bin_union (at level 49).
  Infix "cap" := bin_inter (at level 49).

  Proposition bin_inter_inc_l1 A B : A cap B inc A.
  Proof. apply (inter_inc_l _ (bool_map A B) true). Qed.

  Proposition bin_inter_inc_l2 A B : A cap B inc B.
  Proof. apply (inter_inc_l _ (bool_map A B) false). Qed.

  Hint Immediate bin_inter_inc_l1 bin_inter_inc_l2.

  Proposition bin_inter_inc_l1' A B C : C inc A cap B -> C inc A.
  Proof. intros H; apply inc_transitive with (1:=H); auto. Qed.

  Proposition bin_inter_inc_l2' A B C : C inc A cap B -> C inc B.
  Proof. intros H; apply inc_transitive with (1:=H); auto. Qed.

  Hint Resolve bin_inter_inc_l1' bin_inter_inc_l2'.

  Proposition bin_inter_inc_l A B C : C inc A cap B -> (C inc A /\ C inc B).
  Proof. split; eauto. Qed.
 
  Hint Resolve bin_inter_inc_l.

  Proposition bin_inter_inc_r A B C : C inc A -> C inc B -> C inc A cap B.
  Proof. intros; apply inter_inc_r; intros x; induction x; auto. Qed.

  Hint Resolve bin_inter_inc_r.

  Proposition bin_inter_sg A B x : (A cap B) x -> A x /\ B x.
  Proof. split; apply inc_in_transitive with (B:=A cap B); auto. Qed. 

  Proposition bin_inter_sg' (A B : _ -> Prop) x : A x -> B x -> (A cap B) x.
  Proof. intros; apply in_sg; auto. Qed. 

  Hint Resolve bin_inter_sg bin_inter_sg'.

  Proposition bin_inter_comm A B : A cap B inc B cap A.
  Proof. intros x Hx; apply bin_inter_sg in Hx; destruct Hx; auto. Qed.

  Hint Immediate bin_inter_comm.

  Proposition bin_inter_comm_id A B : A cap B ~ B cap A.
  Proof. split; auto. Qed.

  Hint Immediate bin_inter_comm_id.

  Proposition bin_inter_glb A B C : C inc A cap B <->  (C inc A /\ C inc B).
  Proof. split; auto. intros [ ]; auto. Qed.

  Proposition bin_union_inc_r1 A B : A inc A cup B.
  Proof. apply (union_inc_r bool (bool_map A B) true). Qed. 

  Proposition bin_union_inc_r2 A B : B inc A cup B.
  Proof. apply (union_inc_r bool (bool_map A B) false). Qed.

  Hint Immediate bin_union_inc_r1 bin_union_inc_r2.

  Proposition bin_union_inc_r1' A B C : A cup B inc C -> A inc C.
  Proof. intros H; apply inc_transitive with (2:=H); auto. Qed.

  Proposition bin_union_inc_r2' A B C : A cup B inc C -> B inc C.
  Proof. intros H; apply inc_transitive with (2:=H); auto. Qed.

  Hint Resolve bin_union_inc_r1' bin_union_inc_r2'.

  Proposition bin_union_inc_r A B C : A cup B inc C -> (A inc C /\ B inc C).
  Proof. split; eauto. Qed.

  Proposition bin_union_inc_l A B C : A inc C -> B inc C -> A cup B inc C.
  Proof. intros; apply union_inc_l; intros x; induction x; auto. Qed.

  Hint Resolve bin_union_inc_r bin_union_inc_l.

  Proposition bin_union_sg A B x : (A cup B) x -> A x \/ B x.
  Proof. intros Hx; destruct union_sg with (1 := Hx) as [ [] H ]; simpl in H; tauto. Qed.

  Proposition sg_inc_bin_union x A B : sg x inc A cup B -> (sg x inc A) \/ (sg x inc B).
  Proof. intros H; apply in_sg, bin_union_sg in H; destruct H; [left | right ]; auto. Qed.

End Predicate.

Implicit Arguments Included [ M ].
Implicit Arguments Identical [ M ].
Implicit Arguments Pair [ M ].
Implicit Arguments sg [ M ].
Implicit Arguments inter [M X].
Implicit Arguments union [M X].
Implicit Arguments bin_inter [ M ].
Implicit Arguments bin_union [ M ].

Notation empty := (Emptyset _).
Notation full := (Fullset _).

Infix "inc"   := Included (at level 60, no associativity).
Infix "~"     := Identical (at level 60, no associativity).
Infix "pair"  := Pair (at level 49, no associativity).
Infix "cap"   := bin_inter (at level 49).
Infix "cup"   := bin_union (at level 49).

Hint Immediate inc_reflexive empty_inc_any any_inc_full in_sg_reflexive id_reflexive In_fullset.
Hint Resolve sg_inc in_sg_eq inc_antisym id_symetric.
Hint Immediate inter_inc_l.
Hint Resolve inter_inc_l' inter_inc_r.
Hint Immediate union_inc_r.
Hint Resolve union_inc_r' union_inc_l.
Hint Immediate bin_inter_inc_l1 bin_inter_inc_l2.
Hint Resolve bin_inter_inc_l1' bin_inter_inc_l2'.
Hint Resolve bin_inter_inc_l.
Hint Resolve bin_inter_inc_r.

Hint Resolve bin_inter_sg'.

Hint Immediate bin_union_inc_r1 bin_union_inc_r2.
Hint Resolve bin_union_inc_r1' bin_union_inc_r2'.
Hint Resolve bin_union_inc_r bin_union_inc_l.

Ltac bin_decomp_tac := 
  repeat match goal with 
        [                  |- _ cup _ inc _  ] => apply bin_union_inc_l
      | [                  |- _ inc _ cap _  ] => apply bin_inter_inc_r
      | [ H: (_ cup _) _   |- _              ] => apply bin_union_sg in H; destruct H as [ H | H ]
      | [ H: (_ cap _) _   |- _              ] => apply bin_inter_sg in H; destruct H
      | [ H: _ cup _ inc _ |- _              ] => apply bin_union_inc_r in H; destruct H
      | [ H: _ inc _ cap _ |- _              ] => apply bin_inter_inc_l in H; destruct H
      | [                  |- (_ cap _) _    ] => apply bin_inter_sg'
      | [ H: sg _ _        |- _              ] => apply in_sg_eq in H
      | [                  |- sg _ inc _     ] => apply sg_inc
      | [ H: sg _ inc _    |- _              ] => apply in_sg in H
      | [ H: empty _       |- _              ] => contradiction H
      | [ H: _ /\ _        |- _              ] => destruct H
  end.

Ltac bin_union_inter_tac := 
  match goal with
        [ |- _ inc _ cup _ ] => try (apply inc_transitive with (2 := bin_union_inc_r1 _ _ _); auto; bin_union_inter_tac; fail);
                                try (apply inc_transitive with (2 := bin_union_inc_r2 _ _ _); auto; bin_union_inter_tac; fail)
      | [ |- _ cap _ inc _ ] => try (apply inc_transitive with (1 := bin_inter_inc_l1 _ _ _); auto; bin_union_inter_tac; fail);
                                try (apply inc_transitive with (1 := bin_inter_inc_l2 _ _ _); auto; bin_union_inter_tac; fail)
      | [ |- (_ cup _) _   ] => try (apply bin_union_inc_r1; auto; bin_union_inter_tac; fail);
                                try (apply bin_union_inc_r2; auto; bin_union_inter_tac; fail)
  end.

Ltac pred_solve := bin_decomp_tac; auto; try bin_union_inter_tac.

Section Pred_props.

  Variable X : Type.

  Implicit Type (A B C D : X -> Prop).

  Proposition bin_inter_monotone A A' B B' : A inc A' -> B inc B' -> A cap B inc A' cap B'.
  Proof. intros; pred_solve. Qed.
  
  Hint Resolve bin_inter_monotone.

  Proposition bin_inter_congr A A' B B' : A ~ A' -> B ~ B' -> A cap B ~ A' cap B'.
  Proof. intros [ ] [ ]; split; auto. Qed.

  Proposition bin_union_comm A B : A cup B inc B cup A.
  Proof. pred_solve. Qed. 

  Hint Immediate bin_union_comm.

  Proposition bin_union_comm_id A B : A cup B ~ B cup A.
  Proof. split; auto. Qed.

  Proposition bin_union_lub A B C : A cup B inc C <-> (A inc C /\ B inc C).
  Proof. split; auto. intros [ ]; auto. Qed. 

  Proposition bin_union_monotone A A' B B' : A inc A' -> B inc B' -> A cup B inc A' cup B'.
  Proof. intros; pred_solve. Qed.

  Hint Resolve bin_union_monotone.

  Proposition bin_union_congr A A' B B' : A ~ A' -> B ~ B' -> A cup B ~ A' cup B'.
  Proof. intros [ ] [ ]; split; auto. Qed.
 
  Proposition bin_union_neutral_1 A : A cup empty inc A.
  Proof. auto. Qed.

  Proposition bin_union_neutral_2 A : A inc A cup empty.
  Proof. auto. Qed. 

  Proposition bin_union_neutral A : A cup empty ~ A.
  Proof. split; auto. Qed.

  Proposition bin_union_idem A : A cup A ~ A.
  Proof. split; auto. Qed.

  Proposition bin_union_assoc_1 A B C : (A cup B) cup C inc A cup (B cup C).
  Proof. pred_solve. Qed. 

  Proposition bin_union_assoc_2 A B C : A cup (B cup C) inc (A cup B) cup C.
  Proof. pred_solve. Qed.

  Hint Immediate bin_union_assoc_1 bin_union_assoc_2.

  Proposition bin_union_assoc A B C : A cup (B cup C) ~ (A cup B) cup C.
  Proof. split; auto. Qed.

  Hint Immediate bin_union_assoc.

  Proposition bin_inter_distrib_l1 A B C : (A cup B) cap C inc (A cap C) cup (B cap C).
  Proof. intros ? ?; pred_solve. Qed.

  Proposition bin_inter_distrib_r1 A B C : (A cap C) cup (B cap C) inc (A cup B) cap C.
  Proof. auto. Qed. 

  Hint Immediate bin_inter_distrib_l1 bin_inter_distrib_r1.

  Proposition bin_inter_distrib_left A B C : (A cup B) cap C ~ (A cap C) cup (B cap C).
  Proof. split; auto. Qed.

  Proposition bin_inter_distrib_right A B C : C cap (A cup B) ~ (C cap A) cup (C cap B).
  Proof. split; intros ? ?; pred_solve. Qed.

  Proposition bin_inter_distrib_l2 A B C : C cap (A cup B) inc (C cap A) cup (C cap B).
  Proof. apply bin_inter_distrib_right. Qed.

  Proposition bin_inter_distrib_r2 A B C : (C cap A) cup (C cap B) inc C cap (A cup B).
  Proof. apply bin_inter_distrib_right. Qed.

End Pred_props.

Hint Immediate bin_inter_comm.
Hint Resolve bin_inter_monotone bin_inter_congr.

Hint Immediate bin_union_comm.
Hint Resolve bin_union_monotone bin_union_congr.
Hint Immediate bin_union_assoc_1 bin_union_assoc_2.
Hint Immediate bin_inter_distrib_l1 bin_inter_distrib_r1.
Hint Immediate bin_inter_distrib_l2 bin_inter_distrib_r2.


