(**************************************************************)
(*   The Strong Completeness of Partial Monoidal Boolean BI   *)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import List.

Require Import list_in_t.
Require Import list_fall_exst.
Require Import list_fall_exst_t.

Require Import predicate.
Require Import carrier.

Section Proof_systems.

  Set Implicit Arguments.

  Variable stm : Type. (* stm is the type of propositions/statements of the considered system *)

  Section Rules.

    Record rule : Type := MkRule {
      rparam : Type;
      rpred  : rparam -> Prop;
      rinst  : rparam -> list stm * stm
    }.

    Definition rhyps r p := let (hh,_) := rinst r p in hh.
    Definition rconc r p := let (_ ,c) := rinst r p in c.

    Definition MkZero p cnd c := MkRule (rparam := p) (fun p => cnd p) (fun p => (nil, c p)).
    Definition MkSingle p cnd h1 c := MkRule (rparam := p) (fun p => cnd p) (fun p => ((h1 p)::nil,c p)).
    Definition MkDouble p cnd h1 h2 c := MkRule (rparam := p) (fun p => cnd p) (fun p => ((h1 p)::(h2 p)::nil,c p)).
    Definition MkTriple p cnd h1 h2 h3 c := MkRule (rparam := p) (fun p => cnd p) (fun p => ((h1 p)::(h2 p)::(h3 p)::nil,c p)).

    Definition MkURule p ri := MkRule (rparam := p) (fun _ => True) ri.
    
    Definition MkUAxiom c := MkURule (p := unit) (fun _ => (nil,c)).
    Definition MkUZero p c := MkURule (p := p) (fun p => (nil, c p)).
    Definition MkUSingle p h1 c := MkURule (p := p) (fun p => ((h1 p)::nil,c p)).
    Definition MkUDouble p h1 h2 c := MkURule (p := p) (fun p => ((h1 p)::(h2 p)::nil,c p)).
    Definition MkUTriple p h1 h2 h3 c := MkURule (p := p) (fun p => ((h1 p)::(h2 p)::(h3 p)::nil,c p)).

    Definition MkAxioms P := MkRule P (fun x => (nil,x)).
    
    Definition lstm_to_axioms := map MkUAxiom.

    Definition instance r hh c : Type := { p : rparam r | rpred _ p /\ (hh,c) = rinst _ p }.
    Definition in_instance r hh c p : rpred r p -> hh = rhyps _ p -> c = rconc _ p -> instance r hh c.
    Proof. intros H H1 H2; exists p; split; auto; rewrite H1, H2; unfold rhyps, rconc; destruct (rinst _ p); auto. Qed.

  End Rules.

  Fact rule_hyps_conc_inst r p : (rhyps r p, rconc r p) = rinst r p.
  Proof. unfold rhyps, rconc; destruct (rinst _ p); auto. Qed.

  Section Proofs.

    Variable sys : list rule.

    (* we generate our own induction principles due to nested recursion with lists *)

    Unset Elimination Schemes.
    
    Inductive proof : stm -> Type :=
      | in_proof : forall r hh c, In_t r sys -> instance r hh c -> Forall_t proof hh -> proof c.

    Set Elimination Schemes.

    Section proof_rect.

      Variable   P : stm -> Type.
      Hypothesis Pclosed : forall r hh c, In_t r sys -> instance r hh c -> Forall_t P hh -> P c.

      Fixpoint proof_rect s (p : proof s) := 
        match p with
          | in_proof r hh c H1 H2 H3 => 
            let fall := fix loop hh (ft : Forall_t proof hh) : Forall_t _ hh := 
            match ft in Forall_t _ hh return Forall_t _ hh with
              | in_Forall_t_0 => in_Forall_t_0 _
              | in_Forall_t_1 x ll Hx Hll => in_Forall_t_1 _ (proof_rect Hx) (loop _ Hll)
            end in Pclosed H1 H2 (fall _ H3)
        end.

    End proof_rect.

    Section proof_rec.

      Variable   P : stm -> Set.
      Hypothesis Pclosed : forall r hh c, In_t r sys -> instance r hh c -> Forall_t P hh -> P c.
      Definition proof_rec s (p : proof s) : P s := proof_rect Pclosed p.
   
    End proof_rec.

    Section proof_ind.

      Variable   P       : stm -> Prop.
      Hypothesis Pclosed : forall r hh c, In_t r sys -> instance r hh c -> Forall P hh -> P c.

      Definition proof_ind s (p : proof s) : P s.
      Proof.
        refine (proof_rect _ p).
        intros r hh c Hr I Hhh.
        apply Pclosed with (2 := I); auto.
      Qed.

    End proof_ind.

    Theorem proof_0 r c : In_t r sys -> instance r nil c -> proof c.
    Proof. intros Hr I; apply (in_proof Hr I); auto. Qed.

    Theorem proof_1 r h c : In_t r sys -> instance r (h::nil) c -> proof h -> proof c.
    Proof. intros Hr I Hh; apply (in_proof Hr I); auto. Qed.

    Theorem proof_2 r h1 h2 c : In_t r sys -> instance r (h1::h2::nil) c -> proof h1 -> proof h2 -> proof c.
    Proof. intros Hr I; intros;  apply (in_proof Hr I); auto. Qed.

    Theorem proof_3 r h1 h2 h3 c : In_t r sys -> instance r (h1::h2::h3::nil) c -> proof h1 -> proof h2 -> proof h3 -> proof c.
    Proof. intros Hr I; intros;  apply (in_proof Hr I); auto. Qed.

  End Proofs.

  (* extension of the definition of proof using a (finite) list of axioms *)

  Definition proof_with sys hh c := proof (lstm_to_axioms hh++sys) c.

  Theorem proof_incl sys sys' : In_t_inc sys sys' -> forall p, proof sys p -> proof sys' p.
  Proof.
    intros Inc.
    apply proof_rect with (P := proof sys').
    intros r hh ? ? ? ?;
    apply (in_proof (r := r) (hh := hh)); auto.
  Qed.

  Section extensions.

    Variable sys : list rule.

    (* extension of the definition of proof using a set of axioms *)

    Definition proof_from ax c := proof (MkAxioms ax::sys) c.

    (* the notion of provability and closure under rules *)

    Definition closed_under_rules rel := forall r hh c, In_t r sys -> instance r hh c -> Forall rel hh -> rel c.
    Definition closure_under_proof rel c  := inhabited (proof_from rel c).

    Section closure_ind.

      Variable   rel     : stm -> Prop.
      Variable   P       : stm -> Prop.
      Hypothesis Prel    : rel inc P.
      Hypothesis Pclosed : closed_under_rules P.

      Theorem closure_under_proof_ind : closure_under_proof rel inc P.
      Proof.
        intros c [ H ]. red in H.
        generalize c H; clear c H; apply proof_ind.
        intros r hh c Hr I HH.
        In_t_auto.
        rewrite Hr in I. destruct I as [ p [ H1 H2 ] ].
        simpl in H1, H2.
        injection H2. intros H _. rewrite H; auto.
        exact (Pclosed Hr I HH).
      Qed.
   
    End closure_ind.
        
    Theorem closure_under_proof_is_closed rel : closed_under_rules (closure_under_proof rel).
    Proof.
      intros r hh c S I HH.
      unfold closure_under_proof in HH |- *.
      apply Forall_t_inhabited in HH.
      destruct HH as [ HH ]; constructor.
      unfold proof_from in HH |- *.
      apply in_proof with (r := r) (hh := hh); auto.
      generalize HH; apply Forall_t_map; auto.
    Qed.

    Proposition closed_under_rules_postfixpoint rel : closed_under_rules rel -> closure_under_proof rel inc rel.
    Proof. intro; apply closure_under_proof_ind; auto. Qed.

    Proposition postfixpoint_closed_under_rules rel : closure_under_proof rel inc rel -> closed_under_rules rel.
    Proof.
      intros Hrel r hh c Hr I HH.
      apply Hrel; constructor; red.
      apply in_proof with (r := r) (hh := hh); In_t_auto.
      clear r c Hr I.
      apply In_t_imp_Forall_t.
      intros x Hx.
      apply list_fall_app_0 with (x := x) in HH; auto.
      apply in_proof with (r := MkAxioms rel) (hh := nil); In_t_auto.
      exists x; auto.
    Qed.

    Theorem closure_under_proof_is_inc rel : rel inc closure_under_proof rel.
    Proof.
      intros c Hc.
      constructor.
      red.
      apply proof_0 with (r := MkAxioms rel).
      In_t_auto.
      exists c; auto.
    Qed.

    Hint Immediate closure_under_proof_is_closed closure_under_proof_is_inc.

    Theorem closure_under_proof_is_mono rel1 rel2 : rel1 inc rel2 -> closure_under_proof rel1 inc closure_under_proof rel2.
    Proof.
      intros H.
      apply closure_under_proof_ind; auto.
      apply inc_transitive with (1 := H); auto.
    Qed.

    Hint Resolve closure_under_proof_is_mono.

    Theorem closure_under_proof_is_idem rel : closure_under_proof (closure_under_proof rel) inc closure_under_proof rel.
    Proof.
      apply closure_under_proof_ind with (rel := closure_under_proof rel); auto.
    Qed.

    Hint Immediate closure_under_proof_is_idem.

 
  End extensions.

  Proposition proof_with_from sys ax c : proof_with sys ax c -> proof_from sys (carrier ax) c.
  Proof.
    induction 1. red.
    In_t_auto.

    unfold lstm_to_axioms in X. 
    apply In_t_map_rev in X.
    destruct X as [ h H1 H2 ].
    rewrite H2 in X0. 
    destruct X0 as ( p & _ & H4 ).
    simpl in H4. injection H4; clear H4; intros H4 _.
    rewrite H4.
    apply in_proof with (r := MkAxioms (carrier ax)) (hh := nil); auto.
    exists h; simpl; auto.
    unfold carrier; auto.

    apply in_proof with (r := r) (hh := hh); auto.
    generalize X1; clear X1; apply Forall_t_map.
    intro; auto.
  Qed.
      
End Proof_systems.

Implicit Arguments in_proof [stm sys hh c].
Implicit Arguments proof_0 [stm sys c].
Implicit Arguments proof_1 [stm sys h c].
Implicit Arguments proof_2 [stm sys h1 h2 c].
Implicit Arguments proof_3 [stm sys h1 h2 h3 c].
Implicit Arguments proof_incl [ stm sys sys' ].

Ltac prove_instance k := apply in_instance with (p := k).
Ltac red_instance H p I1 I2 I3 := destruct H as [ p [ I1 I2 ] ]; simpl in p, I1, I2; try (rewrite <- rule_hyps_conc_inst in I2); injection I2; clear I2; intros I2 I3.


